(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
	typeof define === 'function' && define.amd ? define(factory) :
	(global.$ = factory());
}(this, (function () { 'use strict';

(function () {
    if (!Element.prototype.matches) {
        Element.prototype.matches = Element.prototype.matchesSelector || 
        Element.prototype.mozMatchesSelector ||
        Element.prototype.msMatchesSelector || 
        Element.prototype.oMatchesSelector || 
        Element.prototype.webkitMatchesSelector;
    }
})()

/**
 * Filter mixin, keep all filtering methods
 *
 * @class
 */
var Filter = function Filter () {};

Filter.prototype.index = function index () {
    return ( this[0] && this[0].parentNode ) 
        ? this.first().prevAll().length 
        : -1
};

/**
 * Reduce the set of matched elements
 * to the one at the specified index
 *
 * @param {Integer} index
 * @return {jBit} instance
 */
Filter.prototype.at = function at (index) {
    if (index < 0) {
        index = this.length + index;
    }

    return this[index] ? this._make(this[index]) : this._make(null)
};

/**
 * @return {jBit} instance containing the first element
 */
Filter.prototype.first = function first () {
    return this.at(0)
};

/**
 * @return {jBit} instance containing the last element
 */
Filter.prototype.last = function last () {
    return this.at(-1)
};

/**
 * Filter the set of matched elements excluding 
 * the ones that match the filter
 * 
 * @param {String} filter containing a selector expression
 * @return {jBit} instance
 */
Filter.prototype.not = function not (filter) {
    return this.filter(filter, true)
};

/**
 * Filter the set of matched elements to those that have
 * a descendant that matches the selector or Element
 * 
 * @param {String|Element} selector containing a selector expression
 * @return {jBit} instance
 */
Filter.prototype.has = function has (selector) {
    var elems = this._make(selector, this).get();

    var matches = [].filter.call(this, function (el) {
        for (var i = 0; i < elems.length; i++) {
            if (el.contains(elems[i])) {
                return true
            }
        }
    });

    return this._make(matches)
};

/**
 *  Filter the set of matched elements
 *
 * @param {String} selector containing a selector expression
 * @return {jBit} instance
 */
Filter.prototype.filter = function filter (selector, not) {
        var this$1 = this;
        if ( not === void 0 ) not = false;

    return this._make(
        [].filter.call(this, function (el) {
            if (this$1._isStr(selector)) {
                return !not ? this$1.is(selector, el) : !this$1.is(selector, el)
            }
        })
    )
};

/**
 * Siblings mixin, keep all siblings related methods
 *
 * @class
 */
var Siblings = function Siblings () {};

Siblings.prototype.siblings = function siblings () {
    return this._make(
        this._flatten(
            [].map.call(this, function (el) {
                return [].filter.call(
                    el.parentNode.children, function (child) { return el != child; }
                )
            })
        )
    )
};

/**
 * Find the previous element sibling of each element in
 * the current set of elements
 *
 * @return {jBit} instance
 */
Siblings.prototype.prev = function prev () {
    return this._nextPrev('previous')
};

/**
 * Find the next element sibling of each element in
 * the current set of elements
 *
 * @return {jBit} instance
 */
Siblings.prototype.next = function next () {
    return this._nextPrev('next')
};

/**
 * Get all preceding sibling elements of each element in
 * the current set of elements
 * Can receive a selector to filter the matched siblings
 *
 * @param {String} filter containing a selector expression
 * @return {jBit} instance
 */
Siblings.prototype.prevAll = function prevAll (filter) {
    return this._whileDirection('previousElementSibling', filter)
};

/**
 * Get all following sibling elements of each element in
 * the current set of elements
 * Can receive a selector to filter the matched siblings
 *
 * @param {String} filter containing a selector expression
 * @return {jBit} instance
 */
Siblings.prototype.nextAll = function nextAll (filter) {
    return this._whileDirection('nextElementSibling', filter)
};

/**
 * Map each element getting its next/previous sibling
 * 
 * @param {String} direction to go (next|previous)
 * @return {jBit} instance
 */
Siblings.prototype._nextPrev = function _nextPrev (direction) {
    return this._make(
        [].map.call(this, function (el) {
            return el[(direction + "ElementSibling")]
        })
    )
};

/**
 * Siblings mixin, keep all siblings related methods
 *
 * @class
 */
var Utils = function Utils () {};

Utils.prototype._whileDirection = function _whileDirection (direction, filter) {
        var this$1 = this;

    var result = this._flatten(
        [].map.call(this, function (el) {
            var tmp = [];

            while ( (el = el[direction]) && el.nodeType !== 9  )   {
                if (filter) {
                    this$1.is(filter, el) ? tmp.push(el) : '';
                } else {
                    tmp.push(el);
                }
            }

            return tmp
        })
    );

    return this._make(this._unique(result))
};

/**
 * @param {*} value to be tested
 * @return {Boolean} if the value is a string
 */
Utils.prototype._isStr = function _isStr (value) {
    return value && (typeof value === 'string' || value instanceof String)
};

/**
 * @param {*} value to be tested
 * @return {Boolean} if the value is an Element
 */
Utils.prototype._isElement = function _isElement (value) {
    return (value instanceof Element)
};

/**
 * @param {*} value to be tested
 * @return {Boolean} if the value is a Function
 */
Utils.prototype._isFunction = function _isFunction (value) {
    return (value instanceof Function)
};

/**
 * Flatten an array in one level
 * 
 * @param {Array|NodeList} mixed value to be flattened
 * @return {Array} the flattened array
 */
Utils.prototype._flatten = function _flatten (mixed) {
    return [].concat.apply([], mixed)
};

/**
 * @param {Array} mixed to be checked
 * @return {Array} with unique values
 */
Utils.prototype._unique = function _unique (mixed) {
    return mixed.filter(function (el, i) {
        return mixed.indexOf(el) == i
    })
};

/**
 * @param {Array|NodeList} elements
 * @param {Function} cb
 */
Utils.prototype._each = function _each (elements, cb) {
    [].forEach.call(elements, cb);
};

/**
 * Traverse mixin, keep all traversing
 * methods for ancestors and descendants
 *
 * @class
 */
var Traverse = function Traverse () {};

Traverse.prototype.children = function children (filter) {
        var this$1 = this;

    return this._make(
        this._flatten(
            [].map.call(this, function (el) {
                if (filter) {
                    return [].filter.call(
                        el.children, function (c) { return this$1.is(filter, c); }
                    )
                } else {
                    return this$1.slice(el.children)
                }
            })
        )
    )
};

/**
 *  Find descendants of each element in the
 *  current set of elements
 *
 * @param {String} filter containing a selector expression
 * @return {jBit} instance
 */
Traverse.prototype.find = function find (filter) {
        var this$1 = this;

    return this._make(
        this._flatten(
            [].map.call(this, function (el) {
                return this$1.slice(
                    el.querySelectorAll(filter)
                )
            })
        )
    )
};

/**
 * Get the parent elements of each element in
 * the current set of elements
 *
 * @return {jBit} instance
 */
Traverse.prototype.parent = function parent () {
    return this._make(
        [].map.call(this, function (el) { return el.parentNode; })
    )
};

/**
 * @param {String} filter containing a selector expression
 * @return {jBit} instance
 */
Traverse.prototype.parents = function parents (filter) {
    return this._whileDirection('parentNode', filter)
};

/**
 * Returns the closest ancestor of the current element (or the current element itself) 
 * which matches the selector for each of the elements in the current
 * set of elements
 * 
 * @param {String} selector
 * @return {jBit} instance
 */
Traverse.prototype.closest = function closest (selector) {
        var this$1 = this;

    if (!selector || !this._isStr(selector)) {
        return this._make(null)
    }

    var matched = this.map(function (el) {
        //use the native new method if possible
        if (el.closest) {
            return el.closest(selector) 
        } else {
            var parent = el
              , next = null;

            while (
                (next = parent && parent.matches) &&
                !this$1.is(selector, parent)
            ) {
                parent = parent.parentNode;
            }
            return next ? parent : null
        }
    });

    return this._make(matched)
};

/**
 * Regex to match id, tag and class selector
 * 
 * @const
 */
var rgxSelector = /^(?:#([\w-]+)|(\w+)|\.([\w-]+))$/;

/**
 * Add a mixin to a class
 *
 * @function
 * @param  {Object} target class to receive the mixin properties
 * @param  {Object} source mixin class
 */
var mixin = function (target, source) {
    target = target.prototype;
    source = source.prototype;

    Object.getOwnPropertyNames(source).forEach(function (name) {
        if (name !== 'constructor') {
            Object.defineProperty(
                target, name, Object.getOwnPropertyDescriptor(source, name)
            );
        }
    });
};

var Base = function Base () {};

mixin(Base, Filter);
mixin(Base, Siblings);
mixin(Base, Utils);
mixin(Base, Traverse);

/**
 * jBit class
 *
 * @class
 */
var jBit = (function (Base) {
    function jBit (selector, context) {
        if ( context === void 0 ) context = document;

        this.length = 0;

        if (selector instanceof jBit) {
            selector = selector.get();
        }

        if (this._isStr(selector)) {
            if (context instanceof jBit) {
                return context.find(selector)
            }

            this._fill(this._selector(selector, context));
        } else if (Array.isArray(selector)) {
            this._pushArray(selector, context);
        } else if (this._isElement(selector)) {
            this._push(selector);
        }

        return this
    }

    if ( Base ) jBit.__proto__ = Base;
    jBit.prototype = Object.create( Base && Base.prototype );
    jBit.prototype.constructor = jBit;

    /**
     * @param {Function} cb map callback
     * @return {Array} resulted by the map function
     */
    jBit.prototype.map = function map (cb) {
        return this._make(
            [].map.call(this, cb)
        )
    };

    /**
     * @param {Function} cb forEach callback
     */
    jBit.prototype.each = function each (cb) {
        [].forEach.call(this, cb);
    };

    /**
     * @param {Array|NodeList} iterable list
     * @return {Array}
     */
    jBit.prototype.slice = function slice (iterable) {
        return [].slice.apply(iterable, arguments)
    };

    /**
     * Test if each element in the current set of elements
     * match the given selector
     * Can receive an element to compare, ignoring the current
     * set of elements
     *
     * @param {String} selector string containing a selector expression
     * @param {Element} elem to test the selector against
     * @return {jBit} instance
     */
    jBit.prototype.is = function is (selector, elem) {
        if ( elem === void 0 ) elem = null;

        if (elem) {
            return elem.matches(selector)
        } else {
            return [].every.call(this, function (el) { return el.matches(selector); })
        }
    };

    /**
     * Add new elements to the instance
     * 
     * @param {String} selector string containing a selector expression
     * @param {Element} elem to test the selector against
     * @return {jBit} instance
     */
    jBit.prototype.add = function add (selector, context) {
        return this._merge(this, this._make(selector, context))
    };

    /**
     * Add the previous set of elements, to the current one
     * optionally filtered by a selector
     * 
     * @param {String} selector string containing a selector expression
     * @return {jBit} instance
     */
    jBit.prototype.addBack = function addBack (selector) {
        return this.add(
            !selector
                ? this.previous
                : this.previous.filter(selector)
        )
    };

    /**
     * @return {jBit} the previous set of metched elements
     */
    jBit.prototype.end = function end () {
        return this.previous || this._make(null)
    };

    /**
     * Get an array containing the current set
     * of elements
     *
     * @return {Array} array containing the current set of elements
     */
    jBit.prototype.get = function get () {
        return this.slice(this)
    };

    //=================
    //Private methods
    //=================

    /**
     * Used internally to create a new instance for
     * the new set of elements
     * 
     * @param {(String|Array|Element|jBit)} selector
     * @param {Element} context
     * @return {jBit} a new instance
     */
    jBit.prototype._make = function _make (selector, context) {
        var n = new this.constructor(selector, context);

        n.previous = this;

        return n
        // return new this.constructor(selector, context)
    };

    /**
     * @param {(String|Array|Element|jBit)} selector
     * @param {Element} context
     * @return {NodeList} with the matched elements
     */
    jBit.prototype._selector = function _selector (selector, context) {
        var match, m;
        if (match = rgxSelector.exec(selector)) {
            var matched = null;
            //id selector
            if ( (m = match[1]) ) {
                matched = [context.getElementById(m)];
            } else if ( (m = match[2]) ) {
                matched = context.getElementsByTagName(m);
            } else if ( (m = match[3]) ) {
                matched = context.getElementsByClassName(m);
            }

            return matched
        } else {
            return context.querySelectorAll(selector)
        }
    };

    /**
     * Fill the instance with the set of elements
     * 
     * @param {Array} data of elements to fill
     */
    jBit.prototype._fill = function _fill (data) {
        var this$1 = this;

        this._each(data, function (el) { return this$1._push(el); });
    };

    /**
     * Fill an array of selectors|elements|(jBit instances)
     * into the instance
     * 
     * @param {Array} selectorArr Array that may contain strings, elements or jBit instances
     * @param {Element} context
     */
    jBit.prototype._pushArray = function _pushArray (selectorArr, context) {
        var this$1 = this;

        this._fill(
            this._flatten(
                selectorArr.map(function (el) {
                    if (el instanceof jBit) {
                        this$1._pushArray(el.get(), context);
                    }

                    if (this$1._isStr(el)) {
                        return this$1.slice(this$1._selector(el, context))
                    } else if (this$1._isElement(el)) {
                        return el
                    }
                })
            )
        );
    };

    /**
     * Merge two jBit instances
     * 
     * @param {jBit} source instance
     * @param {jBit} target instance
     * @return {jBit} instance containing both, source and target matched elements
     */
    jBit.prototype._merge = function _merge (source, target) {
        for (var i = 0; i < target.length; i++) {
            source._push(target[i]);
        }

        return source
    };

    /**
     * @param {Element} element to be pushed into the instance
     */
    jBit.prototype._push = function _push (element) {
        if (element) {
            [].push.call(this, element);
        }
    };

    return jBit;
}(Base));

/**
 * Wrapper for our jBit class, that allows
 * we to just call $([selector]) without the
 * necessity of instantiating the jBit class
 *
 * @function
 * @param {(String|Array|Element|jBit)} selector
 * @param {Element} context
 * @return {jBit} a new instance
 */
function $ (selector, context) {
    return new jBit(selector, context)
}

var _jQuery = window.jQuery;
var _$ = window.$;

$.noConflict = function () {
	if ( window.$ === jQuery ) {
		window.$ = _$;
	}

	if ( deep && window.jQuery === jQuery ) {
		window.jQuery = _jQuery;
	}

	return jQuery;
};

return $;

})));
//# sourceMappingURL=jbit.js.map
